

#include <fstream>


#include "itkCastImageFilter.h"
#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkNearestNeighborInterpolateImageFunction.h"
#include "itkResampleImageFilter.h"
#include "vnl/vnl_math.h"

#include "itkekADRevolveBSplineDeformableTransform.h"
#include "itkekLoadSaveAffineTransform.h"
#include "itkekADRevolveRBFTransform.h"



int main( int argc, char **argv )
{


	if( argc < 5 || argc > 6)
    {
		std::cerr << "Missing Parameters " << std::endl;
		std::cerr << "Usage:  " << argv[0];
		std::cerr << "  inputImageFile";
		std::cerr << "  outputImagefile";  
		std::cerr << "  referenceImagefile";  
		std::cerr << "  parametersfile";
		std::cerr << "  [typeOfInterpolator]" << std::endl;;  
		return 1;
    }

	char *    inputImageFilename          =      argv[1];
	char *    outputImageFilename         =      argv[2];
	char *    referenceImageFilename      =      argv[3];
	char *    parametersfile              =      argv[4];
	int typeOfInterpolator          = 1;

	if(argc == 6)
		typeOfInterpolator          = atoi(argv[5]);

	const    unsigned int   Dimension = 3;

	typedef  float  PixelType;
  
	typedef itk::Image< PixelType, Dimension >  InputImageType;

	typedef   float     InternalPixelType;

	typedef itk::Image< InternalPixelType, Dimension > InternalImageType;


	typedef itk::ImageFileReader< InputImageType > InputImageReaderType;

	typedef itk::CastImageFilter< 
		InputImageType, InternalImageType > InputCastFilterType;

	InputImageReaderType::Pointer inputImageReader =InputImageReaderType::New();

	InputImageReaderType::Pointer referenceImageReader=InputImageReaderType::New();

	inputImageReader->SetFileName( inputImageFilename );

	referenceImageReader->SetFileName( referenceImageFilename );

	InputCastFilterType::Pointer inputCaster = InputCastFilterType::New();

	inputCaster->SetInput( inputImageReader->GetOutput() );

	inputCaster->Update();

	referenceImageReader->Update();

	InternalImageType::ConstPointer inputImage = inputCaster->GetOutput();

	InputImageType::ConstPointer referenceImage =referenceImageReader->GetOutput();


	typedef itk::LinearInterpolateImageFunction< 
		InternalImageType,
		double             > InterpolatorType1;
    
	typedef itk::NearestNeighborInterpolateImageFunction< 
		InternalImageType,
		double             > InterpolatorType2;


	typedef itk::Transform<double,3,3> TransformType;


	InterpolatorType1::Pointer   interpolator1  = InterpolatorType1::New();
	InterpolatorType2::Pointer   interpolator2  = InterpolatorType2::New();
	TransformType::Pointer      transform; // will be initialized later.


	std::cout.precision(16);
  

	std::string name;

	{
		std::ifstream fin(parametersfile);

		if(!fin)
		{
			std::cout << "Error: could not open file for loading transform" 
					  << std::endl;
			abort();
		}

		// Read the first word of the file which contains the name of the class
		fin >> name;

		fin.close();
	}


	// Set up the transform

	if(name == "ADRevolveBSplineDeformableTransform")
	{

		const unsigned int SplineOrder = 3;
		typedef itkek::ADRevolveBSplineDeformableTransform<
			double,
			Dimension,
			SplineOrder >     TransformType2;
      
		TransformType2::Pointer      transform2     = TransformType2::New();
		transform2->LoadTransform(parametersfile);

		transform = transform2;
	}
	else if(name == "ADRevolveRBFTransform")
	{
		typedef itkek::ADRevolveRBFTransform TransformType2;
		TransformType2::Pointer      transform2     = TransformType2::New();
		transform2->LoadTransform(parametersfile);

		transform = transform2;
	}
	else 
	{
		// If we're not one of the above, then the only other option is that we
		// are an affine transform. The format for an affine transform is just a
		// list of 16 numbers (in ascii). To determine if this is the case, load
		// the values and check if each one is a valid double precision number
		// and also that the number of loaded values is 16.
      
		std::ifstream fin(parametersfile);

		if(!fin)
		{
			std::cout << "Error: could not open file for loading transform" 
					  << std::endl;
			abort();
		}

		int i = 0;
		bool error = false;
		double val;

		while(fin >> val)
		{
			if(++i > 16 || !vnl_math_isfinite(val))
			{
				error = true;
				break;
			}
		}

		if(i != 16)
			error = true;

		fin.close();

		if(error)
		{
			std::cout << "Error reading input file" << std::endl;
			std::cout << "Could not recognize type of transform " << std::endl;
			abort();
		}

		// OK, if we've reached here, we can be sure the transform is affine. So
		// now load it.

		typedef itk::AffineTransform< double, Dimension > TransformType2;

      
		TransformType2::Pointer      transform2     = TransformType2::New();
  

		// Load the transform from a file
		int ret = itkek::LoadSaveAffineTransform::LoadAffineTransform(
			parametersfile,transform2);
		if(ret != 0)
		{
			std::cout << "Error: could not load affine transform. " << std::endl;
			abort();
		}


		transform = transform2;

	}

   
	typedef itk::ResampleImageFilter< 
		InternalImageType, 
		InternalImageType >    ResampleFilterType;


	ResampleFilterType::Pointer resample = ResampleFilterType::New();

  
	resample->SetTransform( transform );


	switch(typeOfInterpolator)
	{
	case 1: 
		resample->SetInterpolator(interpolator1);
		break;  
	case 2: 
		resample->SetInterpolator(interpolator2);
		break;  
	default:
		std::cout << "Unrecognized interpolator." << std::endl;
		abort();
	}


	resample->SetInput( inputImage );

	// Disable multithreading since code is not thread safe
	resample->SetNumberOfThreads(1);
  
	resample->SetSize(    referenceImage->GetLargestPossibleRegion().GetSize() );
	resample->SetOutputOrigin(  referenceImage->GetOrigin() );
	resample->SetOutputSpacing( referenceImage->GetSpacing() );
	resample->SetDefaultPixelValue( 0 );

  
//  typedef  short  OutputPixelType;
	typedef  float  OutputPixelType;

	typedef itk::Image< OutputPixelType, Dimension > OutputImageType;
  
	typedef itk::CastImageFilter< 
		InternalImageType,
		OutputImageType > CastFilterType;
                    
	typedef itk::ImageFileWriter< OutputImageType >  WriterType;


	WriterType::Pointer      writer =  WriterType::New();
	CastFilterType::Pointer  caster =  CastFilterType::New();



	writer->SetFileName( outputImageFilename );
  
  
	caster->SetInput( resample->GetOutput() );
	writer->SetInput( caster->GetOutput()   );

	PRINT("Starting resampling...");
	//std::cout << "Starting resampling..." << std::endl;
	writer->Update();


	return 0;

}

