#ifndef __itkekADRevolveRBFTransform_h
#define __itkekADRevolveRBFTransform_h

#include "itkekADRevolveTransformBase.h"
#include "ObjectiveFunctionBase.h"
#include "itkekRBFMatrix.h"
#include "itkPoint.h"
#include "itkVector.h"
#include "itkPointSet.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkCovariantVector.h"
#include "itkImage.h"
#include "itkVectorLinearInterpolateImageFunction.h"
#include "itkAffineTransform.h"
#include "itkCentralDifferenceImageFunction.h"
#include "vnl/vnl_matrix.h"
#include "vnl/vnl_matrix_fixed.h"


#include <vector>



namespace itkek
{


/**
   This class implements a transformation based on a multiquadric radial basis
   function (which included thin plate spline for tau = 0 and k = 1). 

   The code here for GMRES is from the "Templates" book:  Barrett, R., Berry,
   M., Chan, T.F., Demmel, J., Donato, J., Dongarra, J., Eijkhout, V., Pozo,
   R., Romine, C., and der Vorst, H.V. (1994).  {\em Templates for the
   Solution of Linear Systems: Building Blocks for Iterative Methods, 2nd
   Edition}.  SIAM, Philadelphia, PA.

   Everything in this class is implemented so as to take advantage of
   automatic differentiation with checkpointing. It should also work is a
   normal transform even if automatic differentiation is not used.
 */
class ITK_EXPORT ADRevolveRBFTransform : public ADRevolveTransformBase<double,3,3>
{

   public:

      typedef ADRevolveRBFTransform    Self;
      typedef ADRevolveTransformBase<double,3,3> Superclass;
      
      typedef itk::SmartPointer<Self>         Pointer;
      typedef itk::SmartPointer<const Self>   ConstPointer;


      /** Method for creation through the object factory. */
      itkNewMacro(Self);

      /** Run-time type information (and related methods). */
      itkTypeMacro(ADRevolveRBFTransform,ADRevolveTransformBase);



      typedef itk::Point<double,3> PointType;
      typedef std::vector<PointType> PointsContainer;
      typedef itk::Vector<double,3> VectorType;


      typedef PointType InputPointType;
      typedef PointType OutputPointType;
      typedef itk::Vector<double,3> InputVectorType;
      typedef itk::Vector<double,3> OutputVectorType;

      typedef std::vector<std::pair<int,
				    VectorType> > FixedDisplacementsContainer;

      typedef itk::Image< float,3 > MovingImageType ;
      typedef MovingImageType::ConstPointer MovingImagePointer;
      typedef itk::Image< float,3 > FixedImageType ;
      typedef FixedImageType::ConstPointer FixedImagePointer;

      typedef itk::LinearInterpolateImageFunction<
	 MovingImageType, double > 	InterpolatorType;
      typedef InterpolatorType::Pointer 	MovingImageInterpolatorPointer;

      typedef itk::AffineTransform< double, 3 > AffineTransformType;
      typedef AffineTransformType::Pointer AffineTransformPointer;

      /** The following function must be called before using this class.
	  It is the only setup function */
      void Setup(double tau,
		 int k,
		 int gmresIterations,
		 double gmresTolerance,
		 PointsContainer& sourceLandmarks,
		 AffineTransformPointer affineTransform,
		 bool checkpointingFlag,
		 const FixedDisplacementsContainer& fixedDisplacements = 
		 FixedDisplacementsContainer());

      const PointsContainer& GetSourceLandmarks() const {return s_SourceLandmarks;}


      virtual OutputPointType TransformPoint(const InputPointType& point) const;

      virtual void TransformPointAD( const InputPointType& thisPoint, 
				     ADVectorType& transformedPoint ) const
      {
	    this->TransformPoint__(thisPoint,transformedPoint);
      }

      virtual void Function(adolc::ObjectiveFunctionBase::Label& label);

      virtual adolc::ObjectiveFunctionBase::
      LabelsContainerType GetLabels() const;

      virtual void SetParametersAD(const adouble* ptr) {s_ParametersAD=ptr;}

      virtual void CopyData(const Superclass* rhs);

      virtual void SetParameters( const ParametersType& parameters );

      virtual const ParametersType& GetParameters(void) const;

      virtual unsigned int GetNumberOfParameters(void) const
      {
		  return 3*s_NumberOfPoints;
      }

      void SaveTransform(const char* filename) const;

      void LoadTransform(const char* filename);


      virtual bool DoesTransformHaveMoreThanOneTimeStep() const {return true;}

      //static int ComputeOptimalDepth(int numberOfPoints);


   protected:

      ADRevolveRBFTransform(): Superclass(3,0),m_ParametersAD(0)
      { int numberInts = sizeof(m_Ints) / sizeof(int);
        for(int i=0;i<numberInts;i++)  m_Ints[i] = 0;
      }

      virtual ~ADRevolveRBFTransform() { 
		  if(m_ParametersAD) delete [] m_ParametersAD; }

   private:


      typedef adoublev ADVectorType;
      typedef adoublem ADMatrixType;
      typedef itk::CovariantVector< double,3 >  GradientPixelType;
      typedef itk::Image< GradientPixelType,3 > GradientImageType ;
      typedef itk::SmartPointer< GradientImageType > GradientImagePointer;
      typedef  itk::VectorLinearInterpolateImageFunction<
         GradientImageType, double >    GradientInterpolatorType;
      typedef GradientInterpolatorType::Pointer GradientInterpolatorPointer;
      typedef vnl_matrix_fixed<double,3,4> AffineMatrixType;

      //These variables are static since they don't change 
      // throughout the entire function evaluation
      typedef rbfmatrix<InputPointType> RBFMatrixType;
      typedef std::vector<InputPointType> JacobianSamplePointsContainer;

      static const adouble* s_ParametersAD;
      static double s_Tolerance;
      static RBFMatrixType s_M;
      static int s_NumberOfPoints;
      static int s_MaxGMRESRestartIterations;
      static int s_MaxGMRESIterations;
      static int s_K;
      static double s_Tau;
      static double s_Tau2;
      static PointsContainer s_SourceLandmarks;
      static AffineTransformType::Pointer s_AffineTransform;
      static AffineMatrixType s_AffineMatrix;
      static int s_MatrixMultiplicationBlockSize;
      static FixedDisplacementsContainer s_FixedDisplacements;


      void SetupInterpolationMatrix() const;

      void GMRES(adolc::ObjectiveFunctionBase::Label& label);

      double BasicFunction(const InputPointType& point, 
			   const InputPointType& center) const
      {
	  const double x = point[0] - center[0];
	  const double y = point[1] - center[1];
	  const double z = point[2] - center[2];

	  if(s_K == 1)
	    return std::sqrt( x*x + y*y + z*z + s_Tau2 );
	  else
	    return std::sqrt( std::pow(x*x + y*y + z*z + s_Tau2, s_K ) );
      }

      void UpdateGMRES(ADVectorType& x, const int k, 
		       const ADMatrixType& h, 
		       const ADVectorType& s, 
		       const ADMatrixType& v) const ;

//      void ApplyPlaneRotation(adouble& dx, adouble& dy, 
//			      adouble& cs, adouble& sn) const;
      void ApplyPlaneRotation(badouble dx, badouble dy, 
			      badouble cs, badouble sn) const;

//      void GeneratePlaneRotation(adouble& dx, adouble& dy, 
//				 adouble& cs, adouble& sn) const;
      void GeneratePlaneRotation(badouble dx, badouble dy, 
				 badouble cs, badouble sn) const;

      void Norm(const ADVectorType& v, const int size, 
		adouble& norm) const;

      void DotProduct(const ADVectorType& v1, 
		      const ADVectorType& v2, const int size, 
		      adouble& result) const;

      void TransformPoint__( const InputPointType& thisPoint, 
				ADVectorType& transformedPoint ) const;

      ADRevolveRBFTransform(const ADRevolveRBFTransform& rhs); // not imple
      ADRevolveRBFTransform& operator=(const ADRevolveRBFTransform& rhs); //not imple


      // Static adoubles
      static ADMatrixType* s_X;
//      static ADNonLocalPolynomialCollectionType* s_Poly;

      //Local data
      int m_Ints[15];


      // This object is used to store the parameters ONLY when checkpointing
      // is NOT used
      adouble* m_ParametersAD;
};

}

#endif
